import sys
from datetime import datetime
from typing import List, Optional

if sys.version_info >= (3, 8):
    from typing import TypedDict
else:
    from typing_extensions import TypedDict

from localstack.aws.api import RequestContext, ServiceException, ServiceRequest, handler

Boolean = bool
BoxedBoolean = bool
BoxedDouble = float
Integer = int
ListStatementsLimit = int
Location = str
PageSize = int
ParameterName = str
ParameterValue = str
SecretArn = str
StatementId = str
StatementNameString = str
StatementString = str
String = str
bool = bool


class StatementStatusString(str):
    SUBMITTED = "SUBMITTED"
    PICKED = "PICKED"
    STARTED = "STARTED"
    FINISHED = "FINISHED"
    ABORTED = "ABORTED"
    FAILED = "FAILED"


class StatusString(str):
    SUBMITTED = "SUBMITTED"
    PICKED = "PICKED"
    STARTED = "STARTED"
    FINISHED = "FINISHED"
    ABORTED = "ABORTED"
    FAILED = "FAILED"
    ALL = "ALL"


class ActiveStatementsExceededException(ServiceException):
    """The number of active statements exceeds the limit."""

    Message: Optional[String]


class BatchExecuteStatementException(ServiceException):
    """An SQL statement encountered an environmental error while running."""

    Message: String
    StatementId: String


class DatabaseConnectionException(ServiceException):
    """Connection to a database failed."""

    Message: String


class ExecuteStatementException(ServiceException):
    """The SQL statement encountered an environmental error while running."""

    Message: String
    StatementId: String


class InternalServerException(ServiceException):
    """The Amazon Redshift Data API operation failed due to invalid input."""

    Message: String


class ResourceNotFoundException(ServiceException):
    """The Amazon Redshift Data API operation failed due to a missing resource."""

    Message: String
    ResourceId: String


class ValidationException(ServiceException):
    """The Amazon Redshift Data API operation failed due to invalid input."""

    Message: Optional[String]


SqlList = List[StatementString]


class BatchExecuteStatementInput(ServiceRequest):
    ClusterIdentifier: Optional[Location]
    Database: String
    DbUser: Optional[String]
    SecretArn: Optional[SecretArn]
    Sqls: SqlList
    StatementName: Optional[StatementNameString]
    WithEvent: Optional[Boolean]


Timestamp = datetime


class BatchExecuteStatementOutput(TypedDict, total=False):
    ClusterIdentifier: Optional[Location]
    CreatedAt: Optional[Timestamp]
    Database: Optional[String]
    DbUser: Optional[String]
    Id: Optional[StatementId]
    SecretArn: Optional[SecretArn]


Blob = bytes
BoxedLong = int


class CancelStatementRequest(ServiceRequest):
    Id: StatementId


class CancelStatementResponse(TypedDict, total=False):
    Status: Optional[Boolean]


class ColumnMetadata(TypedDict, total=False):
    """The properties (metadata) of a column."""

    columnDefault: Optional[String]
    isCaseSensitive: Optional[bool]
    isCurrency: Optional[bool]
    isSigned: Optional[bool]
    label: Optional[String]
    length: Optional[Integer]
    name: Optional[String]
    nullable: Optional[Integer]
    precision: Optional[Integer]
    scale: Optional[Integer]
    schemaName: Optional[String]
    tableName: Optional[String]
    typeName: Optional[String]


ColumnList = List[ColumnMetadata]
ColumnMetadataList = List[ColumnMetadata]
DatabaseList = List[String]


class DescribeStatementRequest(ServiceRequest):
    Id: StatementId


Long = int


class SubStatementData(TypedDict, total=False):
    """Information about an SQL statement."""

    CreatedAt: Optional[Timestamp]
    Duration: Optional[Long]
    Error: Optional[String]
    HasResultSet: Optional[Boolean]
    Id: StatementId
    QueryString: Optional[StatementString]
    RedshiftQueryId: Optional[Long]
    ResultRows: Optional[Long]
    ResultSize: Optional[Long]
    Status: Optional[StatementStatusString]
    UpdatedAt: Optional[Timestamp]


SubStatementList = List[SubStatementData]


class SqlParameter(TypedDict, total=False):
    """A parameter used in a SQL statement."""

    name: ParameterName
    value: ParameterValue


SqlParametersList = List[SqlParameter]


class DescribeStatementResponse(TypedDict, total=False):
    ClusterIdentifier: Optional[String]
    CreatedAt: Optional[Timestamp]
    Database: Optional[String]
    DbUser: Optional[String]
    Duration: Optional[Long]
    Error: Optional[String]
    HasResultSet: Optional[Boolean]
    Id: StatementId
    QueryParameters: Optional[SqlParametersList]
    QueryString: Optional[StatementString]
    RedshiftPid: Optional[Long]
    RedshiftQueryId: Optional[Long]
    ResultRows: Optional[Long]
    ResultSize: Optional[Long]
    SecretArn: Optional[SecretArn]
    Status: Optional[StatusString]
    SubStatements: Optional[SubStatementList]
    UpdatedAt: Optional[Timestamp]


class DescribeTableRequest(ServiceRequest):
    ClusterIdentifier: Optional[Location]
    ConnectedDatabase: Optional[String]
    Database: String
    DbUser: Optional[String]
    MaxResults: Optional[PageSize]
    NextToken: Optional[String]
    Schema: Optional[String]
    SecretArn: Optional[SecretArn]
    Table: Optional[String]


class DescribeTableResponse(TypedDict, total=False):
    ColumnList: Optional[ColumnList]
    NextToken: Optional[String]
    TableName: Optional[String]


class ExecuteStatementInput(ServiceRequest):
    ClusterIdentifier: Optional[Location]
    Database: String
    DbUser: Optional[String]
    Parameters: Optional[SqlParametersList]
    SecretArn: Optional[SecretArn]
    Sql: StatementString
    StatementName: Optional[StatementNameString]
    WithEvent: Optional[Boolean]


class ExecuteStatementOutput(TypedDict, total=False):
    ClusterIdentifier: Optional[Location]
    CreatedAt: Optional[Timestamp]
    Database: Optional[String]
    DbUser: Optional[String]
    Id: Optional[StatementId]
    SecretArn: Optional[SecretArn]


class Field(TypedDict, total=False):
    """A data value in a column."""

    blobValue: Optional[Blob]
    booleanValue: Optional[BoxedBoolean]
    doubleValue: Optional[BoxedDouble]
    isNull: Optional[BoxedBoolean]
    longValue: Optional[BoxedLong]
    stringValue: Optional[String]


FieldList = List[Field]


class GetStatementResultRequest(ServiceRequest):
    Id: StatementId
    NextToken: Optional[String]


SqlRecords = List[FieldList]


class GetStatementResultResponse(TypedDict, total=False):
    ColumnMetadata: Optional[ColumnMetadataList]
    NextToken: Optional[String]
    Records: SqlRecords
    TotalNumRows: Optional[Long]


class ListDatabasesRequest(ServiceRequest):
    ClusterIdentifier: Optional[Location]
    Database: String
    DbUser: Optional[String]
    MaxResults: Optional[PageSize]
    NextToken: Optional[String]
    SecretArn: Optional[SecretArn]


class ListDatabasesResponse(TypedDict, total=False):
    Databases: Optional[DatabaseList]
    NextToken: Optional[String]


class ListSchemasRequest(ServiceRequest):
    ClusterIdentifier: Optional[Location]
    ConnectedDatabase: Optional[String]
    Database: String
    DbUser: Optional[String]
    MaxResults: Optional[PageSize]
    NextToken: Optional[String]
    SchemaPattern: Optional[String]
    SecretArn: Optional[SecretArn]


SchemaList = List[String]


class ListSchemasResponse(TypedDict, total=False):
    NextToken: Optional[String]
    Schemas: Optional[SchemaList]


class ListStatementsRequest(ServiceRequest):
    MaxResults: Optional[ListStatementsLimit]
    NextToken: Optional[String]
    RoleLevel: Optional[Boolean]
    StatementName: Optional[StatementNameString]
    Status: Optional[StatusString]


StatementStringList = List[StatementString]


class StatementData(TypedDict, total=False):
    """The SQL statement to run."""

    CreatedAt: Optional[Timestamp]
    Id: StatementId
    IsBatchStatement: Optional[Boolean]
    QueryParameters: Optional[SqlParametersList]
    QueryString: Optional[StatementString]
    QueryStrings: Optional[StatementStringList]
    SecretArn: Optional[SecretArn]
    StatementName: Optional[StatementNameString]
    Status: Optional[StatusString]
    UpdatedAt: Optional[Timestamp]


StatementList = List[StatementData]


class ListStatementsResponse(TypedDict, total=False):
    NextToken: Optional[String]
    Statements: StatementList


class ListTablesRequest(ServiceRequest):
    ClusterIdentifier: Optional[Location]
    ConnectedDatabase: Optional[String]
    Database: String
    DbUser: Optional[String]
    MaxResults: Optional[PageSize]
    NextToken: Optional[String]
    SchemaPattern: Optional[String]
    SecretArn: Optional[SecretArn]
    TablePattern: Optional[String]


TableMember = TypedDict(
    "TableMember",
    {
        "name": Optional[String],
        "schema": Optional[String],
        "type": Optional[String],
    },
    total=False,
)
TableList = List[TableMember]


class ListTablesResponse(TypedDict, total=False):
    NextToken: Optional[String]
    Tables: Optional[TableList]


class RedshiftDataApi:

    service = "redshift-data"
    version = "2019-12-20"

    @handler("BatchExecuteStatement")
    def batch_execute_statement(
        self,
        context: RequestContext,
        database: String,
        sqls: SqlList,
        cluster_identifier: Location = None,
        db_user: String = None,
        secret_arn: SecretArn = None,
        statement_name: StatementNameString = None,
        with_event: Boolean = None,
    ) -> BatchExecuteStatementOutput:
        """Runs one or more SQL statements, which can be data manipulation language
        (DML) or data definition language (DDL). Depending on the authorization
        method, use one of the following combinations of request parameters:

        -  Secrets Manager - when connecting to a cluster, specify the Amazon
           Resource Name (ARN) of the secret, the database name, and the cluster
           identifier that matches the cluster in the secret. When connecting to
           a serverless endpoint, specify the Amazon Resource Name (ARN) of the
           secret and the database name.

        -  Temporary credentials - when connecting to a cluster, specify the
           cluster identifier, the database name, and the database user name.
           Also, permission to call the ``redshift:GetClusterCredentials``
           operation is required. When connecting to a serverless endpoint,
           specify the database name.

        :param database: The name of the database.
        :param sqls: One or more SQL statements to run.
        :param cluster_identifier: The cluster identifier.
        :param db_user: The database user name.
        :param secret_arn: The name or ARN of the secret that enables access to the database.
        :param statement_name: The name of the SQL statements.
        :param with_event: A value that indicates whether to send an event to the Amazon
        EventBridge event bus after the SQL statements run.
        :returns: BatchExecuteStatementOutput
        :raises ValidationException:
        :raises ActiveStatementsExceededException:
        :raises BatchExecuteStatementException:
        """
        raise NotImplementedError

    @handler("CancelStatement")
    def cancel_statement(self, context: RequestContext, id: StatementId) -> CancelStatementResponse:
        """Cancels a running query. To be canceled, a query must be running.

        :param id: The identifier of the SQL statement to cancel.
        :returns: CancelStatementResponse
        :raises ValidationException:
        :raises ResourceNotFoundException:
        :raises InternalServerException:
        :raises DatabaseConnectionException:
        """
        raise NotImplementedError

    @handler("DescribeStatement")
    def describe_statement(
        self, context: RequestContext, id: StatementId
    ) -> DescribeStatementResponse:
        """Describes the details about a specific instance when a query was run by
        the Amazon Redshift Data API. The information includes when the query
        started, when it finished, the query status, the number of rows
        returned, and the SQL statement.

        :param id: The identifier of the SQL statement to describe.
        :returns: DescribeStatementResponse
        :raises ValidationException:
        :raises ResourceNotFoundException:
        :raises InternalServerException:
        """
        raise NotImplementedError

    @handler("DescribeTable")
    def describe_table(
        self,
        context: RequestContext,
        database: String,
        cluster_identifier: Location = None,
        connected_database: String = None,
        db_user: String = None,
        max_results: PageSize = None,
        next_token: String = None,
        schema: String = None,
        secret_arn: SecretArn = None,
        table: String = None,
    ) -> DescribeTableResponse:
        """Describes the detailed information about a table from metadata in the
        cluster. The information includes its columns. A token is returned to
        page through the column list. Depending on the authorization method, use
        one of the following combinations of request parameters:

        -  Secrets Manager - when connecting to a cluster, specify the Amazon
           Resource Name (ARN) of the secret, the database name, and the cluster
           identifier that matches the cluster in the secret. When connecting to
           a serverless endpoint, specify the Amazon Resource Name (ARN) of the
           secret and the database name.

        -  Temporary credentials - when connecting to a cluster, specify the
           cluster identifier, the database name, and the database user name.
           Also, permission to call the ``redshift:GetClusterCredentials``
           operation is required. When connecting to a serverless endpoint,
           specify the database name.

        :param database: The name of the database that contains the tables to be described.
        :param cluster_identifier: The cluster identifier.
        :param connected_database: A database name.
        :param db_user: The database user name.
        :param max_results: The maximum number of tables to return in the response.
        :param next_token: A value that indicates the starting point for the next set of response
        records in a subsequent request.
        :param schema: The schema that contains the table.
        :param secret_arn: The name or ARN of the secret that enables access to the database.
        :param table: The table name.
        :returns: DescribeTableResponse
        :raises ValidationException:
        :raises InternalServerException:
        :raises DatabaseConnectionException:
        """
        raise NotImplementedError

    @handler("ExecuteStatement")
    def execute_statement(
        self,
        context: RequestContext,
        database: String,
        sql: StatementString,
        cluster_identifier: Location = None,
        db_user: String = None,
        parameters: SqlParametersList = None,
        secret_arn: SecretArn = None,
        statement_name: StatementNameString = None,
        with_event: Boolean = None,
    ) -> ExecuteStatementOutput:
        """Runs an SQL statement, which can be data manipulation language (DML) or
        data definition language (DDL). This statement must be a single SQL
        statement. Depending on the authorization method, use one of the
        following combinations of request parameters:

        -  Secrets Manager - when connecting to a cluster, specify the Amazon
           Resource Name (ARN) of the secret, the database name, and the cluster
           identifier that matches the cluster in the secret. When connecting to
           a serverless endpoint, specify the Amazon Resource Name (ARN) of the
           secret and the database name.

        -  Temporary credentials - when connecting to a cluster, specify the
           cluster identifier, the database name, and the database user name.
           Also, permission to call the ``redshift:GetClusterCredentials``
           operation is required. When connecting to a serverless endpoint,
           specify the database name.

        :param database: The name of the database.
        :param sql: The SQL statement text to run.
        :param cluster_identifier: The cluster identifier.
        :param db_user: The database user name.
        :param parameters: The parameters for the SQL statement.
        :param secret_arn: The name or ARN of the secret that enables access to the database.
        :param statement_name: The name of the SQL statement.
        :param with_event: A value that indicates whether to send an event to the Amazon
        EventBridge event bus after the SQL statement runs.
        :returns: ExecuteStatementOutput
        :raises ValidationException:
        :raises ExecuteStatementException:
        :raises ActiveStatementsExceededException:
        """
        raise NotImplementedError

    @handler("GetStatementResult")
    def get_statement_result(
        self, context: RequestContext, id: StatementId, next_token: String = None
    ) -> GetStatementResultResponse:
        """Fetches the temporarily cached result of an SQL statement. A token is
        returned to page through the statement results.

        :param id: The identifier of the SQL statement whose results are to be fetched.
        :param next_token: A value that indicates the starting point for the next set of response
        records in a subsequent request.
        :returns: GetStatementResultResponse
        :raises ValidationException:
        :raises ResourceNotFoundException:
        :raises InternalServerException:
        """
        raise NotImplementedError

    @handler("ListDatabases")
    def list_databases(
        self,
        context: RequestContext,
        database: String,
        cluster_identifier: Location = None,
        db_user: String = None,
        max_results: PageSize = None,
        next_token: String = None,
        secret_arn: SecretArn = None,
    ) -> ListDatabasesResponse:
        """List the databases in a cluster. A token is returned to page through the
        database list. Depending on the authorization method, use one of the
        following combinations of request parameters:

        -  Secrets Manager - when connecting to a cluster, specify the Amazon
           Resource Name (ARN) of the secret, the database name, and the cluster
           identifier that matches the cluster in the secret. When connecting to
           a serverless endpoint, specify the Amazon Resource Name (ARN) of the
           secret and the database name.

        -  Temporary credentials - when connecting to a cluster, specify the
           cluster identifier, the database name, and the database user name.
           Also, permission to call the ``redshift:GetClusterCredentials``
           operation is required. When connecting to a serverless endpoint,
           specify the database name.

        :param database: The name of the database.
        :param cluster_identifier: The cluster identifier.
        :param db_user: The database user name.
        :param max_results: The maximum number of databases to return in the response.
        :param next_token: A value that indicates the starting point for the next set of response
        records in a subsequent request.
        :param secret_arn: The name or ARN of the secret that enables access to the database.
        :returns: ListDatabasesResponse
        :raises ValidationException:
        :raises InternalServerException:
        :raises DatabaseConnectionException:
        """
        raise NotImplementedError

    @handler("ListSchemas")
    def list_schemas(
        self,
        context: RequestContext,
        database: String,
        cluster_identifier: Location = None,
        connected_database: String = None,
        db_user: String = None,
        max_results: PageSize = None,
        next_token: String = None,
        schema_pattern: String = None,
        secret_arn: SecretArn = None,
    ) -> ListSchemasResponse:
        """Lists the schemas in a database. A token is returned to page through the
        schema list. Depending on the authorization method, use one of the
        following combinations of request parameters:

        -  Secrets Manager - when connecting to a cluster, specify the Amazon
           Resource Name (ARN) of the secret, the database name, and the cluster
           identifier that matches the cluster in the secret. When connecting to
           a serverless endpoint, specify the Amazon Resource Name (ARN) of the
           secret and the database name.

        -  Temporary credentials - when connecting to a cluster, specify the
           cluster identifier, the database name, and the database user name.
           Also, permission to call the ``redshift:GetClusterCredentials``
           operation is required. When connecting to a serverless endpoint,
           specify the database name.

        :param database: The name of the database that contains the schemas to list.
        :param cluster_identifier: The cluster identifier.
        :param connected_database: A database name.
        :param db_user: The database user name.
        :param max_results: The maximum number of schemas to return in the response.
        :param next_token: A value that indicates the starting point for the next set of response
        records in a subsequent request.
        :param schema_pattern: A pattern to filter results by schema name.
        :param secret_arn: The name or ARN of the secret that enables access to the database.
        :returns: ListSchemasResponse
        :raises ValidationException:
        :raises InternalServerException:
        :raises DatabaseConnectionException:
        """
        raise NotImplementedError

    @handler("ListStatements")
    def list_statements(
        self,
        context: RequestContext,
        max_results: ListStatementsLimit = None,
        next_token: String = None,
        role_level: Boolean = None,
        statement_name: StatementNameString = None,
        status: StatusString = None,
    ) -> ListStatementsResponse:
        """List of SQL statements. By default, only finished statements are shown.
        A token is returned to page through the statement list.

        :param max_results: The maximum number of SQL statements to return in the response.
        :param next_token: A value that indicates the starting point for the next set of response
        records in a subsequent request.
        :param role_level: A value that filters which statements to return in the response.
        :param statement_name: The name of the SQL statement specified as input to
        ``BatchExecuteStatement`` or ``ExecuteStatement`` to identify the query.
        :param status: The status of the SQL statement to list.
        :returns: ListStatementsResponse
        :raises ValidationException:
        :raises InternalServerException:
        """
        raise NotImplementedError

    @handler("ListTables")
    def list_tables(
        self,
        context: RequestContext,
        database: String,
        cluster_identifier: Location = None,
        connected_database: String = None,
        db_user: String = None,
        max_results: PageSize = None,
        next_token: String = None,
        schema_pattern: String = None,
        secret_arn: SecretArn = None,
        table_pattern: String = None,
    ) -> ListTablesResponse:
        """List the tables in a database. If neither ``SchemaPattern`` nor
        ``TablePattern`` are specified, then all tables in the database are
        returned. A token is returned to page through the table list. Depending
        on the authorization method, use one of the following combinations of
        request parameters:

        -  Secrets Manager - when connecting to a cluster, specify the Amazon
           Resource Name (ARN) of the secret, the database name, and the cluster
           identifier that matches the cluster in the secret. When connecting to
           a serverless endpoint, specify the Amazon Resource Name (ARN) of the
           secret and the database name.

        -  Temporary credentials - when connecting to a cluster, specify the
           cluster identifier, the database name, and the database user name.
           Also, permission to call the ``redshift:GetClusterCredentials``
           operation is required. When connecting to a serverless endpoint,
           specify the database name.

        :param database: The name of the database that contains the tables to list.
        :param cluster_identifier: The cluster identifier.
        :param connected_database: A database name.
        :param db_user: The database user name.
        :param max_results: The maximum number of tables to return in the response.
        :param next_token: A value that indicates the starting point for the next set of response
        records in a subsequent request.
        :param schema_pattern: A pattern to filter results by schema name.
        :param secret_arn: The name or ARN of the secret that enables access to the database.
        :param table_pattern: A pattern to filter results by table name.
        :returns: ListTablesResponse
        :raises ValidationException:
        :raises InternalServerException:
        :raises DatabaseConnectionException:
        """
        raise NotImplementedError



_M='state_archive'
_L='presigned_url_metadata'
_K='presigned_url_state'
_J='max_ver'
_I='cpvcs'
_H='ci_pod'
_G='qualifying_name'
_F='presigned_urls'
_E='pod_name'
_D='storage_uuid'
_C=True
_B=None
_A=False
import json,logging,os,re
from abc import ABCMeta,abstractmethod
from typing import Dict,List,Optional,Set,Union
from zipfile import ZipFile
import requests
from localstack import config,constants
from localstack.utils.common import clone,cp_r,disk_usage,download,load_file,new_tmp_dir,new_tmp_file,retry,rm_rf,safe_requests,save_file,to_str
from localstack.utils.docker_utils import DOCKER_CLIENT
from localstack.utils.testutil import create_zip_file
from requests import Response
from localstack_ext.bootstrap.licensing import get_auth_headers
from localstack_ext.bootstrap.pods import pods_api
from localstack_ext.bootstrap.pods.models import Serialization
from localstack_ext.bootstrap.pods.pods_api import PodsApi
from localstack_ext.bootstrap.state_utils import API_STATES_DIR,DYNAMODB_DIR,KINESIS_DIR,api_states_traverse
from localstack_ext.constants import API_PATH_PODS
LOG=logging.getLogger(__name__)
PERSISTED_FOLDERS=[API_STATES_DIR,DYNAMODB_DIR,KINESIS_DIR]
class PodInfo:
	def __init__(A,name=_B,pod_size=0):A.name=name;A.pod_size=pod_size;A.pod_size_compressed=0;A.persisted_resource_names=[]
class CloudPodsManager(metaclass=ABCMeta):
	@abstractmethod
	def __init__(self,pod_name,pre_config):A=self;A.pod_name=pod_name;A.pod_config=clone(pre_config);A.pods_api=PodsApi()
	@abstractmethod
	def init(self):...
	@abstractmethod
	def delete(self,remote):...
	@abstractmethod
	def push(self,comment=_B):...
	@abstractmethod
	def push_overwrite(self,version,comment=_B):...
	@abstractmethod
	def pull(self,inject_version_state=_A,reset_state_before=_A,lazy=_A):...
	@abstractmethod
	def commit(self,message):...
	@abstractmethod
	def inject(self,version,reset_state):...
	@abstractmethod
	def list_versions(self):...
	@abstractmethod
	def version_info(self,version):...
	@abstractmethod
	def version_metamodel(self,version):...
	@abstractmethod
	def set_version(self,version,inject_version_state,reset_state,commit_before):...
	@abstractmethod
	def list_version_commits(self,version):...
	@abstractmethod
	def get_commit_diff(self,version,commit):...
	@abstractmethod
	def register_remote(self,pod_name,ci_pod):...
	@abstractmethod
	def rename_pod(self,current_pod_name,new_pod_name):...
	@abstractmethod
	def list_pods(self,fetch_remote):...
	@staticmethod
	def restart_container():
		LOG.info('Restarting LocalStack instance with updated persistence state - this may take some time ...');B={'action':'restart'};A='%s/health'%config.get_edge_url()
		try:requests.post(A,data=json.dumps(B))
		except requests.exceptions.ConnectionError:pass
		def C():LOG.info('Waiting for LocalStack instance to be fully initialized ...');B=requests.get(A);C=json.loads(to_str(B.content));D=[A for(B,A)in C['services'].items()];assert set(D)=={'running'}
		retry(C,sleep=3,retries=10)
	@staticmethod
	def get_state_zip_from_instance(get_content=_A):
		B=f"{get_pods_endpoint()}/state";A=requests.get(B)
		if A.status_code>=400:raise Exception('Unable to get local pod state via management API %s (code %s): %s'%(B,A.status_code,A.content))
		if get_content:return A.content
		C=f"{new_tmp_file()}.zip";save_file(C,A.content);return C
	def get_pod_info(C,pod_data_dir=_B):
		A=pod_data_dir;B=PodInfo(C.pod_name)
		if A:B.pod_size=disk_usage(A);B.persisted_resource_names=get_persisted_resource_names(A)
		return B
class CloudPodsVersionManager(CloudPodsManager):
	def __init__(A,pod_name,pre_config):super().__init__(pod_name,pre_config)
	@staticmethod
	def parse_pod_name_from_qualifying_name(qualifying_name):return qualifying_name.split(PODS_NAMESPACE_DELIM,1)[1]
	@staticmethod
	def _prepare_archives_from_pre_signed_urls(content):
		A=new_tmp_file();B=content.get(_F);I=B.get('presigned_version_space_url');download(url=I,path=A);C={};D={};J=B.get('presigned_meta_state_urls')
		for (E,F) in J.items():G=new_tmp_file();H=new_tmp_file();K=F['meta'];L=F['state'];download(K,G);download(L,H);D[E]=G;C[E]=H
		return A,D,C
	@classmethod
	def _get_max_version_for_pod_from_platform(B,pod_name,auth_headers):
		C=CloudPodsVersionManager.create_platform_url(f"{pod_name}/info/max-version");A=safe_requests.get(url=C,headers=auth_headers);D='Failed to get version information from platform.. aborting'
		if not B._check_response(A,message=D):return
		E=json.loads(A.content);F=int(E[_J]);return F
	def _add_state_files_func(B,**A):C=A.get('dir_name');D=A.get('account_id');E=A.get('fname');F=A.get('region');G=A.get('service_name');B.pods_api.create_state_file_from_fs(path=C,file_name=E,service=G,region=F,account_id=D,root=API_STATES_DIR,serialization=Serialization.MAIN)
	def _add_state_files_from_directory(B,service,path,root_dir,serialization):
		A=path
		if not os.path.isdir(A):return
		for C in os.listdir(A):B.pods_api.create_state_file_from_fs(path=A,file_name=C,service=service,region='NA',root=root_dir,serialization=serialization)
	def _add_state_to_cloud_pods_store(B,extract_assets=_A):
		if not B.pods_api.config_context.is_initialized():LOG.debug('No Cloud Pod instance detected in the local context - unable to push');return
		C=B.get_state_zip_from_instance();A=new_tmp_dir()
		with ZipFile(C,'r')as G:
			G.extractall(A);H=os.path.join(A,API_STATES_DIR);api_states_traverse(api_states_path=H,side_effect=B._add_state_files_func,mutables=_B);I=os.path.join(A,KINESIS_DIR);J=os.path.join(A,DYNAMODB_DIR);B._add_state_files_from_directory(service='kinesis',path=I,root_dir=KINESIS_DIR,serialization=Serialization.KINESIS);B._add_state_files_from_directory(service='dynamodb',path=J,root_dir=DYNAMODB_DIR,serialization=Serialization.DDB)
			if extract_assets:
				D=[]
				for E in os.listdir(A):
					F=os.path.join(A,E)
					if os.path.isdir(F)and E not in PERSISTED_FOLDERS:D.append(F)
				B.pods_api.add_assets_to_pod(assets_root_paths=D)
		rm_rf(C);rm_rf(A)
	def _pull_versions(A,auth_headers,required_versions):
		D=A.create_platform_url(f"{A.pod_name}?versions={required_versions}");C=safe_requests.get(url=D,headers=auth_headers);E='Failed to pull requested versions from platform'
		if not A._check_response(C,message=E):return
		F=json.loads(C.content);B=CloudPodsVersionManager._prepare_archives_from_pre_signed_urls(F);G=B[0];H=B[1];I=B[2];A.pods_api.merge_from_remote(version_space_archive=G,meta_archives=H,state_archives=I)
	def _clone_pod(A,auth_headers,lazy=_A):
		D=A.create_platform_url(f"{A.pod_name}/clone")
		if lazy:D+='?lazy=True'
		E=safe_requests.get(D,headers=auth_headers);G=f"Failed to clone requested pod {A.pod_name}: <content>"
		if not A._check_response(E,message=G):return
		B=json.loads(E.content);C=CloudPodsVersionManager._prepare_archives_from_pre_signed_urls(B);H=C[0];I=C[1];J=C[2];F={_D:B.get(_D),_G:B.get(_E)};K=CloudPodsVersionManager.parse_pod_name_from_qualifying_name(F[_G]);A.pods_api.init_remote(pod_name=K,version_space_archive=H,meta_archives=I,state_archives=J,remote_info=F)
	def init(A):A.pods_api.init(pod_name=A.pod_name)
	def delete(A,remote):
		C=A.pods_api.config_context.cloud_pods_root_dir;B=os.path.join(C,A.pod_name)
		if os.path.isdir(B):rm_rf(B);return _C
		if remote:0
		return _A
	def _push_to_remote(A,url):
		C=get_auth_headers();B=safe_requests.put(url=url,headers=C);D='Failed to get presigned URLs to upload new version.. aborting'
		if not A._check_response(B,message=D):return
		E=json.loads(B.content);F=E.get(_F);A.pods_api.upload_version_and_product_space(presigned_urls=F)
	def push(A,comment=_B):
		B=comment;A.pods_api.set_pod_context(A.pod_name);A._add_state_to_cloud_pods_store(extract_assets=_C)
		if A.pods_api.is_remotely_managed():
			D=get_auth_headers();C=A.pods_api.get_max_version_no();E=A._get_max_version_for_pod_from_platform(pod_name=A.pod_name,auth_headers=D)
			if C<E:A.pull()
			A.pods_api.push(comment=B);F=A.create_platform_url(f"push/{A.pod_name}?version={C+1}");A._push_to_remote(url=F)
		else:G=A.pods_api.push(comment=B);LOG.debug(f"Created new version: {G}")
		return PodInfo()
	def push_overwrite(A,version,comment=_B):
		B=version;A.pods_api.set_pod_context(pod_name=A.pod_name)
		if B>A.pods_api.get_max_version_no():LOG.warning(f"Version {B} does not exist");return _A
		A._add_state_to_cloud_pods_store();A.pods_api.push_overwrite(version=B,comment=comment)
		if A.pods_api.is_remotely_managed():C=CloudPodsVersionManager.create_platform_url(f"push-overwrite/{A.pod_name}?version={B}");A._push_to_remote(url=C)
		return _C
	def pull(A,inject_version_state=_A,reset_state_before=_A,lazy=_A):
		B=get_auth_headers()
		if A.pod_name in A.pods_api.list_locally_available_pods(show_remote_or_local=_A):
			A.pods_api.set_pod_context(A.pod_name);C=CloudPodsVersionManager._get_max_version_for_pod_from_platform(A.pod_name,B)
			if not C:return
			D=A.pods_api.get_max_version_no()
			if C==D:LOG.info('No new version available remotely. Nothing to pull');return
			if not lazy:E=','.join(map(lambda ver:str(ver),range(D+1,C+1)))
			else:E=D
			A._pull_versions(auth_headers=B,required_versions=E)
		else:A._clone_pod(auth_headers=B,lazy=lazy)
	def commit(A,message=_B):A.pods_api.set_pod_context(A.pod_name);A._add_state_to_cloud_pods_store();B=A.pods_api.commit(message=message);LOG.debug(f"Completed revision: {B}")
	def _download_version_product(D,version,retain=_A):
		A=version;E=D._get_presigned_url_for_version_product(version=A);F=E.get(_K);G=E.get(_L);B=new_tmp_file();C=new_tmp_file();download(F,B);download(G,C)
		if retain:from localstack_ext.bootstrap.pods.utils.remote_utils import extract_meta_and_state_archives as H;H(meta_archives={A:C},state_archives={A:B},config_context=D.pods_api.config_context)
		return{'metadata_archive':C,_M:B}
	def _get_presigned_url_for_version_product(A,version):
		B=version;C=A.create_platform_url(f"{A.pod_name}/version/product")
		if B!=-1:C+=f"?version={B}"
		E=get_auth_headers();D=safe_requests.get(C,headers=E);F=f"Failed to retrieve presigned URL from remote for version {B} of pod {A.pod_name}"
		if not A._check_response(D,message=F):return
		return json.loads(D.content)
	def _inject_from_remote(B,version,retain=_A):
		C=version;from localstack_ext.bootstrap.pods.utils.remote_utils import extract_meta_and_state_archives as F;D=B._get_presigned_url_for_version_product(version=C);G=D.get(_K);A=new_tmp_file();download(G,A);B.deploy_pod_into_instance(pod_path=A)
		if not retain:rm_rf(A);return _C
		H=D.get(_L);E=new_tmp_file();download(H,E);F(meta_archives={C:E},state_archives={C:A},config_context=B.pods_api.config_context);return _C
	@staticmethod
	def deploy_pod_into_instance(pod_path):
		A=pod_path;D=_A
		if os.path.isdir(A):
			B=new_tmp_dir()
			for E in PERSISTED_FOLDERS:
				F=os.path.join(A,E)
				if not os.path.exists(F):continue
				H=os.path.join(B,E);cp_r(F,H,rm_dest_on_conflict=_C)
			A=create_zip_file(B);rm_rf(B);D=_C
		I=load_file(A,mode='rb');G=get_pods_endpoint();C=requests.post(G,data=I)
		if C.status_code>=400:raise Exception('Unable to restore pod state via local pods management API %s (code %s): %s'%(G,C.status_code,C.content))
		if D:rm_rf(A)
		else:return A
	def inject(A,version,reset_state):
		B=version
		if not A.pods_api.config_context.pod_exists_locally(A.pod_name):LOG.debug(f"Pod {A.pod_name} does not exist locally. Requesting state from remote..");C=A._download_version_product(version=B).get(_M)
		else:
			A.pods_api.set_pod_context(A.pod_name)
			if B==-1:B=A.pods_api.get_max_version_no()
			C=A.pods_api.config_context.get_version_state_archive(B)
			if not C and A.pods_api.is_remotely_managed():
				LOG.debug('Fetching requested archive from remote..');D=A._download_version_product(version=B,retain=_C)
				if not D:return _A
				C=A.pods_api.commit_metamodel_utils.get_version_state_archive(B)
		if reset_state:reset_local_state(reset_data_dir=_C)
		A.deploy_pod_into_instance(C);return _C
	def list_versions(A):A.pods_api.set_pod_context(A.pod_name);B=A.pods_api.list_versions();return B
	def version_info(A,version):
		B=version;A.pods_api.set_pod_context(A.pod_name)
		if B==-1:B=A.pods_api.get_max_version_no()
		C=A.pods_api.get_version_info(B);return C
	def version_metamodel(A,version):
		B=version;from localstack_ext.bootstrap.pods.object_storage import get_object_storage_provider as D;A.pods_api.set_pod_context(A.pod_name)
		if B==-1:B=A.pods_api.get_max_version_no()
		E=A.pods_api.get_version_by_number(B);F=D(A.pods_api.config_context);G=F.get_revision_by_key(E.incoming_revision_ptr);C=A.pods_api.commit_metamodel_utils.reconstruct_metamodel(version=B,revision=G.revision_number+1)
		if not C and A.pods_api.is_remotely_managed():A._download_version_product(version=B,retain=_C);C=A.pods_api.commit_metamodel_utils.create_metamodel_from_state_files(version=B)
		return C
	def set_version(A,version,inject_version_state,reset_state,commit_before):
		B=version;A.pods_api.set_pod_context(A.pod_name);C=A.pods_api.set_active_version(version_no=B,commit_before=commit_before)
		if not C:LOG.warning(f"Could not find version {B}")
		if inject_version_state:A.inject(version=B,reset_state=reset_state)
		return C
	def list_version_commits(A,version):A.pods_api.set_pod_context(A.pod_name);B=A.pods_api.list_version_commits(version_no=version);return B
	def get_commit_diff(A,version,commit):A.pods_api.set_pod_context(A.pod_name);B=A.pods_api.commit_metamodel_utils.get_commit_diff(version_no=version,commit_no=commit);return B
	def register_remote(A,pod_name,ci_pod=_A):
		A.pods_api.set_pod_context(A.pod_name);B=A.pods_api.get_max_version_no()
		if B==0:A.pods_api.push('Init Version');B=1
		F=get_auth_headers();G=A.create_platform_url('register');C={_E:A.pod_name,_J:B,_H:ci_pod};C=json.dumps(C);E=safe_requests.post(G,C,headers=F);H=f"Failed to register pod {A.pod_name}: <content>"
		if not A._check_response(E,message=H):return _A
		D=json.loads(E.content);I={_D:D.get(_D),_G:D.get(_E)};J=D.get(_F);A.pods_api.upload_version_and_product_space(presigned_urls=J);pods_api.register_remote(remote_info=I,config_context=A.pods_api.config_context);return _C
	def rename_pod(A,current_pod_name,new_pod_name):
		C=current_pod_name;B=new_pod_name;A.pods_api.set_pod_context(C)
		if B in A.pods_api.list_locally_available_pods():LOG.warning(f"{B} already exists locally");return _A
		if A.pods_api.is_remotely_managed():
			E=get_auth_headers();F=A.create_platform_url(f"{C}/rename");D={'new_pod_name':B};D=json.dumps(D);G=safe_requests.put(F,D,headers=E);H=f"Failed to rename {C} to {B}: <content>"
			if not A._check_response(G,message=H):return _A
		A.pods_api.rename_pod(B);return _C
	def list_pods(A,fetch_remote):
		B=A.pods_api.list_locally_available_pods()
		if fetch_remote:
			D=get_auth_headers();E=A.create_platform_url('pods');C=safe_requests.get(E,headers=D);A._check_response(C,message='Error fetching list of pods from API (status <status_code>)',raise_error=_C);F=json.loads(C.content)
			for G in F.get('registered_pods')or[]:B.add(f"remote/{G}")
		return B
	@classmethod
	def _check_response(C,response,message,raise_error=_A):
		B=response;A=message
		if B.ok:return _C
		if B.status_code in[401,403]:raise Exception('Access denied - please log in first.')
		A=A.replace('<content>',to_str(B.content));A=A.replace('<status_code>',str(B.status_code));LOG.warning(A)
		if raise_error:raise Exception(A)
		return _A
	@staticmethod
	def create_platform_url(path):A=path;B=f"{constants.API_ENDPOINT}/cpvcs";A=A if A.startswith('/')else f"/{A}";return f"{B}{A}"
class PodConfigManagerMeta(type):
	def __getattr__(C,attr):
		def A(*D,**E):
			A=_B
			for F in C.CHAIN:
				try:
					B=getattr(F,attr)(*(D),**E)
					if B:
						if not A:A=B
						elif isinstance(B,list)and isinstance(A,list):A.extend(B)
				except Exception:
					if LOG.isEnabledFor(logging.DEBUG):LOG.exception('error during PodConfigManager call chain')
			if A is not _B:return A
			raise Exception('Unable to run operation "%s" for local or remote configuration'%attr)
		return A
class PodConfigManager(metaclass=PodConfigManagerMeta):
	CHAIN=[]
	@classmethod
	def pod_config(D,pod_name):
		A=pod_name;C=PodConfigManager.list_pods();B=[B for B in C if B[_E]==A]
		if not B:raise Exception('Unable to find config for pod named "%s"'%A)
		return B[0]
ALL_MANAGERS={_I:CloudPodsVersionManager}
def get_pods_manager(pods_name,pre_config=_B):
	B='backend';A=pre_config
	if A is _B:A={B:_I}
	C=A.get(B,_I)
	try:D=ALL_MANAGERS[C](pod_name=pods_name,pre_config=A)
	except KeyError as E:raise NotImplementedError from E
	return D
def init_cloudpods(pod_name,pre_config,**B):A=get_pods_manager(pods_name=pod_name);A.init()
def delete_pod(pod_name,remote,pre_config):A=get_pods_manager(pods_name=pod_name);B=A.delete(remote=remote);return B
def register_remote(pod_name,pre_config,**D):A=pod_name;B=get_pods_manager(pods_name=A);C=B.register_remote(pod_name=A,ci_pod=pre_config.get(_H,_A));return C
def rename_pod(current_pod_name,new_pod_name,pre_config,**D):A=new_pod_name;B=get_pods_manager(pods_name=A);C=B.rename_pod(current_pod_name=current_pod_name,new_pod_name=A);return C
def list_pods(remote,pre_config,**C):A=get_pods_manager(pods_name='');B=A.list_pods(fetch_remote=remote);return B
def commit_state(pod_name,pre_config,message=_B,**C):
	B=pod_name;A=get_pods_manager(pods_name=B);A.pods_api.set_pod_context(pod_name=B)
	if not A.pods_api.config_context.is_initialized():A.init()
	A.commit(message=message)
def inject_state(pod_name,version,reset_state,pre_config,**C):A=get_pods_manager(pods_name=pod_name);B=A.inject(version=version,reset_state=reset_state);return B
def list_versions(pod_name,pre_config,**C):A=get_pods_manager(pods_name=pod_name);B=A.list_versions();return B
def get_version_info(version,pod_name,pre_config,**C):A=get_pods_manager(pods_name=pod_name);B=A.version_info(version=version);return B
def get_version_metamodel(version,pod_name,pre_config,**C):A=get_pods_manager(pods_name=pod_name);B=A.version_metamodel(version=version);return B
def set_version(version,inject_version_state,reset_state,commit_before,pod_name,pre_config,**C):A=get_pods_manager(pods_name=pod_name);B=A.set_version(version=version,inject_version_state=inject_version_state,reset_state=reset_state,commit_before=commit_before);return B
def list_version_commits(version,pod_name,pre_config):A=get_pods_manager(pods_name=pod_name);B=A.list_version_commits(version=version);return B
def get_commit_diff(version,commit,pod_name,pre_config):A=get_pods_manager(pods_name=pod_name);B=A.get_commit_diff(version=version,commit=commit);return B
def push_overwrite(version,pod_name,comment,pre_config):A=get_pods_manager(pods_name=pod_name);A.push_overwrite(version=version,comment=comment)
def push_state(pod_name,pre_config=_B,comment=_B,register=_A,**F):
	B=pod_name;A=get_pods_manager(pods_name=B);A.pods_api.set_pod_context(pod_name=B)
	if not A.pods_api.config_context.is_initialized():A.init()
	D=clone(A.pod_config);C=A.push(comment=comment);D['size']=C.pod_size or C.pod_size_compressed;D['available_resources']=C.persisted_resource_names;E=_C
	if register:E&=A.register_remote(pod_name=B,ci_pod=pre_config.get(_H,_A))
	return E
def get_pods_endpoint():A=config.get_edge_url();return f"{A}{API_PATH_PODS}"
def pull_state(pod_name,inject_version_state=_A,reset_state_before=_A,lazy=_A,**C):
	A=pod_name
	if not A:raise Exception('Need to specify a pod name')
	B=get_pods_manager(pods_name=A);B.pull(inject_version_state=inject_version_state,reset_state_before=reset_state_before,lazy=lazy);print('Done.')
def reset_local_state(reset_data_dir=_A,exclude_from_reset=_B):
	C=exclude_from_reset;A=f"{get_pods_endpoint()}/state"
	if reset_data_dir:A+='/datadir'
	if C:A+=f"?exclude={','.join(C)}"
	print('Sending request to reset the service states in local instance ...');B=requests.delete(A)
	if B.status_code>=400:raise Exception('Unable to reset service state via local management API %s (code %s): %s'%(A,B.status_code,B.content))
	print('Done.')
def save_pods_config(options):A=get_pods_manager('');A.pods_api.config_context.save_pods_config(options=options)
def get_pods_config():A=get_pods_manager('');return A.pods_api.config_context.get_pods_config_cache()
def is_initialized(pod_name):A=get_pods_manager(pods_name=pod_name);return A.pods_api.config_context.is_initialized()
def get_data_dir_from_container():
	try:
		C=DOCKER_CLIENT.inspect_container(config.MAIN_CONTAINER_NAME);D=C.get('Mounts');E=C.get('Config',{}).get('Env',[]);A=[A for A in E if A.startswith('DATA_DIR=')][0].partition('=')[2]
		try:B=[B for B in D if B['Destination']==A][0]['Source'];B=re.sub('^(/host_mnt)?','',B);A=B
		except Exception:LOG.debug(f"No docker volume for data dir '{A}' detected")
		return A
	except Exception:LOG.warning('Unable to determine DATA_DIR from LocalStack Docker container - please make sure $MAIN_CONTAINER_NAME is configured properly')
def get_persisted_resource_names(data_dir):
	E='api_states';D=data_dir;B=[]
	with os.scandir(D)as C:
		for A in C:
			if A.is_dir()and A.name!=E:B.append(A.name)
	with os.scandir(os.path.join(D,E))as C:
		for A in C:
			if A.is_dir()and len(os.listdir(A.path))>0:B.append(A.name)
	LOG.debug(f"Detected state files for the following APIs: {B}");return B
PODS_NAMESPACE_DELIM='-'


_R='Mosquitto'
_Q='cd %s; git checkout %s; ./bootstrap -DREGRESS_CHECKS=OFF; cd build; make; make install'
_P='cd /tmp; git clone https://github.com/timescale/timescaledb.git'
_O='2.0.0-rc4'
_N='timescaledb'
_M='mysqld'
_L='MariaDB'
_K='PostgreSQL'
_J='plpython3'
_I='postgres11'
_H='Error while installing package %s.'
_G='mosquitto'
_F='redis-server'
_E='install'
_D='/tmp/timescaledb'
_C='gcc'
_B=True
_A=None
import logging,os,threading
from abc import ABC
from functools import lru_cache
from typing import List,Optional,Union
from urllib.parse import urlparse
from localstack import config
from localstack.services.install import download_and_extract_with_retry
from localstack.utils import common
from localstack.utils.collections import ensure_list
from localstack.utils.common import in_docker,is_command_available,is_debian,rm_rf,run
from localstack.utils.files import mkdir
INSTALL_LOCK=threading.RLock()
POSTGRES_RPM_REPOSITORY='https://download.postgresql.org/pub/repos/yum/reporpms/EL-8-x86_64/pgdg-redhat-repo-latest.noarch.rpm'
LOG=logging.getLogger(__name__)
@lru_cache()
def is_redhat():return'rhel'in common.load_file('/etc/os-release','')
class PackageInstallationException(Exception):0
class SystemNotSupportedException(PackageInstallationException):0
class PackageInstaller(ABC):
	def __init__(A,log_package_name):
		B=log_package_name;A.log_package_name=B;A.logger=logging.getLogger(f"{__name__}.{B}")
		if B is _A:A.logger.setLevel(logging.CRITICAL+1)
	def install(A,raise_on_error=_B):
		try:
			if not A._check_if_available():
				try:A.logger.debug('Preparing the installation of %s.',A.log_package_name);A._prepare_installation();A.logger.debug('Starting to install %s.',A.log_package_name);A._install_package();A.logger.debug('Executing post-processing of %s.',A.log_package_name);A._post_process()
				except Exception as B:
					if isinstance(B,PackageInstallationException):raise
					else:C=f" ({A.log_package_name})."if A.log_package_name is not _A else'.';raise PackageInstallationException(f"The installation failed{C}") from B
				A._verify_installation();A.logger.debug('Successfully installed %s.',A.log_package_name)
			else:A.logger.debug('%s is already available.',A.log_package_name)
		except PackageInstallationException as B:
			if raise_on_error:raise
			elif A.logger.isEnabledFor(logging.DEBUG):A.logger.exception(_H,A.log_package_name)
			else:A.logger.error(B)
	def _check_if_available(A):raise NotImplementedError
	def _prepare_installation(A):0
	def _install_package(A):raise NotImplementedError
	def _post_process(A):0
	def _verify_installation(A):
		if not A._check_if_available():B=f" of {A.log_package_name}"if A.log_package_name is not _A else'';raise PackageInstallationException(f"The installation{B} failed (verification failed).")
class MultiPackageInstaller(PackageInstaller):
	def __init__(B,log_package_name,package_installers):A=package_installers;super(MultiPackageInstaller,B).__init__(log_package_name=log_package_name);B.package_installers=A if isinstance(A,List)else[A]
	def install(A,raise_on_error=_B):
		for B in A.package_installers:B.install(raise_on_error=raise_on_error)
class OSSpecificPackageInstaller(MultiPackageInstaller):
	def __init__(B,debian_installers,redhat_installers):
		A=[]
		if is_debian():A=debian_installers
		elif is_redhat():A=redhat_installers
		super(OSSpecificPackageInstaller,B).__init__(log_package_name=B.__class__.__name__,package_installers=A)
	def install(A,raise_on_error=_B):
		B=raise_on_error
		try:
			if not in_docker():raise SystemNotSupportedException('OS level packages are only installed within docker containers.')
			elif not is_debian()and not is_redhat():raise SystemNotSupportedException('The current operating system is currently not supported.')
			else:super(OSSpecificPackageInstaller,A).install(raise_on_error=B)
		except PackageInstallationException as C:
			if B:raise
			elif A.logger.isEnabledFor(logging.DEBUG):A.logger.exception(_H,A.log_package_name)
			else:A.logger.error(C)
class DebianPackageInstaller(PackageInstaller,ABC):
	def __init__(A,package_name):super(DebianPackageInstaller,A).__init__(log_package_name=package_name)
	def _install_os_packages(B,packages):
		A=packages;A=A if isinstance(A,List)else[A]
		with INSTALL_LOCK:C=B._download_os_packages(A);D=B.cmd_prefix(C)+[_E]+A;run(D)
	def _download_os_packages(C,packages):
		A=packages;A=ensure_list(A)
		with INSTALL_LOCK:
			D='--'.join(A);B=C.get_download_path(D);mkdir(B);run(['apt-get','update']);LOG.debug('Downloading packages %s to folder: %s',A,B)
			try:E=C.cmd_prefix(B)+['-d',_E]+A;run(E)
			except Exception as F:LOG.info('Unable to download packages %s: %s',A,F);rm_rf(B)
			return B
	def cmd_prefix(B,cache_dir):A=cache_dir;return['apt',f"-o=dir::cache={A}",f"-o=dir::cache::archives={A}",'-y']
	def get_download_path(A,package):return os.path.join(config.dirs.var_libs,'apt-libs',package)
class RedHatPackageInstaller(PackageInstaller,ABC):
	def __init__(A,package_name):super(RedHatPackageInstaller,A).__init__(log_package_name=package_name)
	def _install_os_packages(B,packages):
		A=packages;A=A if isinstance(A,List)else[A]
		with INSTALL_LOCK:run(['dnf',_E,'-y']+A)
DEBIAN_POSTGRES_LIB_FOLDER='/usr/lib/postgresql/11/lib'
REDHAT_POSTGRES_LIB_FOLDER='/usr/pgsql-11/lib'
class DebianPostgres11Installer(DebianPackageInstaller):
	def __init__(A):super(DebianPostgres11Installer,A).__init__(_I)
	def _check_if_available(A):return is_command_available('psql')
	def _install_package(A):A._install_os_packages('postgresql-11')
class RedHatPostgres11Installer(RedHatPackageInstaller):
	def __init__(A):super(RedHatPostgres11Installer,A).__init__(_I)
	def _check_if_available(A):return is_command_available('psql')
	def _prepare_installation(A):A._install_os_packages(POSTGRES_RPM_REPOSITORY)
	def _install_package(A):A._install_os_packages(['postgresql11-devel','postgresql11-server'])
	def _post_process(A):run('ln -s /usr/pgsql-11/bin/pg_config /usr/bin/pg_config')
class DebianPlPythonInstaller(DebianPackageInstaller):
	def __init__(A):super(DebianPlPythonInstaller,A).__init__(_J)
	def _check_if_available(A):return os.path.exists(f"{DEBIAN_POSTGRES_LIB_FOLDER}/plpython3.so")
	def _install_package(A):A._install_os_packages('postgresql-plpython3-11')
class RedHatPlPythonInstaller(RedHatPackageInstaller):
	def __init__(A):super(RedHatPlPythonInstaller,A).__init__(_J)
	def _check_if_available(A):return os.path.exists(f"{REDHAT_POSTGRES_LIB_FOLDER}/plpython3.so")
	def _install_package(A):A._install_os_packages('postgresql11-plpython3')
postgres_installer=OSSpecificPackageInstaller(debian_installers=MultiPackageInstaller(_K,[DebianPostgres11Installer(),DebianPlPythonInstaller()]),redhat_installers=MultiPackageInstaller(_K,[RedHatPostgres11Installer(),RedHatPlPythonInstaller()]))
class DebianMariaDBInstaller(DebianPackageInstaller):
	def __init__(A):super(DebianMariaDBInstaller,A).__init__(_L)
	def _check_if_available(A):return is_command_available(_M)
	def _install_package(A):A._install_os_packages(['mariadb-server','mariadb-client'])
class RedHatMariaDBInstaller(RedHatPackageInstaller):
	def __init__(A):super(RedHatMariaDBInstaller,A).__init__(_L)
	def _check_if_available(A):return is_command_available(_M)
	def _install_package(A):raise PackageInstallationException('MariaDB currently cannot be installed on RedHat')
mariadb_installer=OSSpecificPackageInstaller(debian_installers=DebianMariaDBInstaller(),redhat_installers=RedHatMariaDBInstaller())
class DebianTimescaleDBInstaller(DebianPackageInstaller):
	def __init__(A):super(DebianTimescaleDBInstaller,A).__init__(_N)
	def _check_if_available(A):return os.path.exists(f"{DEBIAN_POSTGRES_LIB_FOLDER}/timescaledb.so")
	def _install_package(A):A._install_os_packages([_C,'cmake',_C,'git']);B=_D;C=_O;run(_P);run(_Q%(B,C));rm_rf(_D)
class RedHatTimescaleDBInstaller(RedHatPackageInstaller):
	def __init__(A):super(RedHatTimescaleDBInstaller,A).__init__(_N)
	def _check_if_available(A):return os.path.exists(f"{REDHAT_POSTGRES_LIB_FOLDER}/timescaledb.so")
	def _install_package(A):A._install_os_packages([_C,'cmake',_C,'git','redhat-rpm-config']);B=_D;C=_O;run(_P);run(_Q%(B,C));rm_rf(_D)
timescaledb_installer=OSSpecificPackageInstaller(debian_installers=DebianTimescaleDBInstaller(),redhat_installers=RedHatTimescaleDBInstaller())
class DebianRedisInstaller(DebianPackageInstaller):
	def __init__(A):super(DebianRedisInstaller,A).__init__('Redis')
	def _check_if_available(A):return is_command_available(_F)
	def _install_package(A):A._install_os_packages(_F)
class RedHatRedisInstaller(RedHatPackageInstaller):
	def __init__(A):super(RedHatRedisInstaller,A).__init__('Redis')
	def _check_if_available(A):return is_command_available(_F)
	def _install_package(A):raise PackageInstallationException('Redis currently cannot be installed on RedHat')
redis_installer=OSSpecificPackageInstaller(debian_installers=DebianRedisInstaller(),redhat_installers=RedHatRedisInstaller())
class DebianMosquittoInstaller(DebianPackageInstaller):
	def __init__(A):super(DebianMosquittoInstaller,A).__init__(_R)
	def _check_if_available(A):return is_command_available(_G)
	def _install_package(A):A._install_os_packages(_G)
class RedHatMosquittoInstaller(RedHatPackageInstaller):
	def __init__(A):super(RedHatMosquittoInstaller,A).__init__(_R)
	def _check_if_available(A):return is_command_available(_G)
	def _install_package(A):raise PackageInstallationException('Mosquitto currently cannot be installed on RedHat')
mosquitto_installer=OSSpecificPackageInstaller(debian_installers=DebianMosquittoInstaller(),redhat_installers=RedHatMosquittoInstaller())
class ArchiveDownloadInstaller(PackageInstaller):
	def __init__(A,archive_url,target_dir,cache_archive=_A,extract_single_directory=_B):
		C=cache_archive;B=archive_url;super(ArchiveDownloadInstaller,A).__init__(log_package_name=A.__class__.__name__);A.archive_url=B;A.target_dir=target_dir;A.extract_single_directory=extract_single_directory;A.cache_archive=C
		if not C:D=urlparse(B).path.split('.');E=D[-1]if len(D)>1 else'zip';A.cache_archive=f"{A.target_dir.rstrip(os.sep)}.{E}"
	def _check_if_available(A):return os.path.exists(A.target_dir)and bool(os.listdir(A.target_dir))
	def _install_package(A):
		with INSTALL_LOCK:
			A.logger.debug('Downloading and installing archive: %s',A.archive_url);download_and_extract_with_retry(A.archive_url,A.cache_archive,A.target_dir)
			if A.extract_single_directory:
				B=os.listdir(A.target_dir)
				if len(B)!=1:return
				C=os.path.join(A.target_dir,B[0])
				if not os.path.isdir(C):return
				os.rename(C,f"{A.target_dir}.bk");rm_rf(A.target_dir);os.rename(f"{A.target_dir}.bk",A.target_dir)


from typing import Dict,NamedTuple,Union
Blob=bytes
class ServiceKey(NamedTuple):account_id:str;region:str;service:str
Backends=Dict[str,Blob]
class BackendState:
	key:0;backends:0
	def __init__(A,key,backends):A.key=key;A.backends=backends
class ServiceState:
	state:Dict[ServiceKey,BackendState];assets:Dict[str,Blob]
	def __init__(A):A.state={};A.assets={}
	def add(B,state_to_add):
		A=state_to_add
		if isinstance(A,B.__class__):B._merge_with_service(A)
		elif isinstance(A,BackendState):B._add_backend(A)
	def add_asset(A,service,content):A.assets[service]=content
	def _add_backend(B,backend_state):
		A=backend_state
		if A.key not in B.state:B.state[A.key]=A
		else:
			from localstack_ext.bootstrap.pods.utils.merge_utils import get_merge_manager as D;C=D(service=A.key.service)
			if C:C.merge(a=B.state[A.key].backends,b=A.backends)
	def _merge_with_service(B,service):
		A=service
		for C in A.state.values():B._add_backend(C)
		if A.assets:B._merge_assets(A.assets)
	def _merge_assets(A,assets):A.assets.update(assets)
	def is_empty(A):return len(A.state)==0
	def __str__(A):return f"Backends: {A.state.__str__()}\nAssets: {A.assets.__str__()}"


from typing import Type
from localstack.aws.api import CommonServiceException,RequestContext
from localstack.aws.api.ec2 import CopyImageRequest,CopyImageResult,CreateImageRequest,CreateImageResult,DescribeImagesRequest,DescribeImagesResult,DescribeInstancesRequest,DescribeInstancesResult,ImportImageRequest,ImportImageResult,InstanceState,RegisterImageRequest,RegisterImageResult,Reservation,RunInstancesRequest,StartInstancesRequest,StartInstancesResult,StopInstancesRequest,StopInstancesResult,TerminateInstancesRequest,TerminateInstancesResult
from localstack.utils.objects import SubtypesInstanceManager
from localstack_ext import config as config_ext
class InstanceStateCode(int):pending=0;running=16;shutting_down=32;terminated=48;stopped=80
class InternalError(CommonServiceException):
	def __init__(A,message):super(InternalError,A).__init__(code='InternalError',message=message)
class IncorrectInstanceStateError(CommonServiceException):
	def __init__(A,instance_id):super(IncorrectInstanceStateError,A).__init__(code='IncorrectInstanceState',message=f"The instance '{instance_id}' is not in a state from which it can be started")
class InvalidAMIIdError(CommonServiceException):
	def __init__(A,ami_id):super(InvalidAMIIdError,A).__init__(code='InvalidAMIID.NotFound',message=f"The image id '{ami_id}' does not exist")
class InvalidInstanceIdError(CommonServiceException):
	def __init__(A,instance_id):super(InvalidInstanceIdError,A).__init__(code='InvalidInstanceID.NotFound',message=f"The instance ID '{instance_id}' does not exist")
class MissingParameterError(CommonServiceException):
	def __init__(A,parameter):super(MissingParameterError,A).__init__(code='MissingParameter',message=f"The request must contain the parameter {parameter}")
class VmManager(SubtypesInstanceManager):
	@classmethod
	def get_base_type(A):return VmManager
	@classmethod
	def get_manager(A):return A.get(config_ext.EC2_VM_MANAGER)
	def start_instances(A,context,start_instances_request):raise NotImplementedError
	def run_instances(A,context,run_instances_request):raise NotImplementedError
	def stop_instances(A,context,stop_instances_request):raise NotImplementedError
	def terminate_instances(A,context,terminate_instances_request):raise NotImplementedError
	def describe_instances(A,context,describe_instances_request):raise NotImplementedError
	def create_image(A,context,create_image_request):raise NotImplementedError
	def describe_images(A,context,describe_images_request):raise NotImplementedError
	def import_image(A,context,import_image_request):raise NotImplementedError
	def instance_state(A,instance_id):raise NotImplementedError
	def initialise_images(A):raise NotImplementedError
	def register_image(A,context,register_image_request):raise NotImplementedError
	def copy_image(A,context,copy_image_request):raise NotImplementedError


from __future__ import annotations
_C=True
_B=False
_A=None
import inspect,json,logging,os
from copy import deepcopy
from typing import Any,Dict,Optional,Set,Type
from localstack.utils.common import ArbitraryAccessObj
from moto.s3.models import FakeBucket
from moto.sqs.models import Queue
from localstack_ext.bootstrap.state_utils import check_already_visited,check_already_visited_obj_id,get_object_dict
LOG=logging.getLogger(__name__)
def _merge_3way_dfs_on_key(current,injecting,ancestor,key,visited=_A):
	F=visited;D=ancestor;C=injecting;B=current;A=key;G=A in B;H=A in C;I=A in D if D else _B;K=I and not H and G;L=not I and H and not G;J=H and G and B[A]!=C[A];M=J and I and B[A]==D[A]
	if K:del B[A]
	elif L:B[A]=deepcopy(C[A])
	else:
		E=_A
		if M:E=C[A];B[A]=deepcopy(E)
		elif J:E=B[A]
		N,F=check_already_visited_obj_id(E,F)
		if not N:O=B[A];merge_3way(O,C.get(A,{}),D.get(A,{}),F)
def merge_3way(current,injecting,ancestor,visited=_A):
	B=injecting;A=current
	if isinstance(A,list):A.extend(B);return
	elif isinstance(A,set):A.update(B);return
	C=get_object_dict(A)
	if C is not _A:
		D=get_object_dict(B)or{};E=get_object_dict(ancestor)or{};F={*(C),*(D),*(E)}
		for G in F:_merge_3way_dfs_on_key(C,D,E,G,visited)
def merge_object_state(current,injecting,common_ancestor=_A):
	B=injecting;A=current
	if not A or not B:return A
	C=handle_special_case(A,B)
	if C:return A
	merge_3way(A,B,common_ancestor);add_missing_attributes(A);return A
def handle_special_case(current,injecting):
	B=current;A=injecting
	if isinstance(A,Queue):B.queues[A.name]=A;return _C
	elif isinstance(A,FakeBucket):C=B['global']if isinstance(B,dict)else B;C.buckets[A.name]=A;return _C
def add_missing_attributes(obj,safe=_C,visited=_A):
	C=visited;A=obj
	try:
		B=get_object_dict(A)
		if B is _A:return
		E,C=check_already_visited(A,C)
		if E:return
		for F in B.values():add_missing_attributes(F,safe=safe,visited=C)
		G=infer_class_attributes(type(A))
		for (D,H) in G.items():
			if D not in B:LOG.debug("Add missing attribute '%s' to state object of type %s"%(D,type(A)));B[D]=H
	except Exception as I:
		if not safe:raise
		LOG.warning('Unable to add missing attributes to persistence state object %s: %s',(A,I))
def infer_class_attributes(clazz):
	B=clazz
	if B in[list,dict]or not inspect.isclass(B)or B.__name__=='function':return{}
	C=getattr(B,'__init__',_A)
	if not C:return{}
	try:
		A=inspect.getfullargspec(C)
		def D(arg_name,arg_index=-1):
			C=arg_name;B=A.defaults or[];F=len(A.args or[])-len(B);D=arg_index-F
			if D in range(len(B)):return B[D]
			E=A.kwonlydefaults or{}
			if C in E:return E[C]
			return ArbitraryAccessObj()
		E=[];F={}
		for G in range(1,len(A.args)):E.append(D(A.args[G],arg_index=G))
		for H in A.kwonlyargs:F[H]=D(H)
		I=B(*(E),**F);J=dict(I.__dict__);return J
	except Exception:return{}
def merge_kinesis_state(path_dest,path_src):
	J='streams';E=path_src;D=path_dest;F='kinesis-data.json';B=os.path.join(D,F);G=os.path.join(E,F)
	if not os.path.isfile(B):LOG.info(f"Could not find statefile in path destination {D}");return _B
	if not os.path.isfile(G):LOG.info(f"Could not find statefile in path source {E}");return _B
	with open(B)as K,open(G)as L:
		H=json.load(K);M=json.load(L);I=H.get(J,[]);C=M.get(J,[])
		if len(C)>0:
			N=I.keys()
			for A in C:
				if A not in N:I[A]=C.get(A);LOG.debug(f"Copied state from stream {A}")
			with open(B,'w')as O:O.write(json.dumps(H))
			return _C
	return _B


_A='utf-8'
import hashlib,logging,os,random,zipfile
from typing import Optional
from localstack.utils.common import new_tmp_dir,rm_rf
from localstack_ext.bootstrap.pods.models import Revision,Version
from localstack_ext.bootstrap.state_utils import API_STATES_DIR,api_states_traverse
LOG=logging.getLogger(__name__)
def random_hash():return hex(random.getrandbits(160))
def compute_file_hash(file_path,accum=None):
	B=accum;A=file_path
	try:
		with open(A,'rb')as C:
			if B:B.update(C.read())
			else:return hashlib.sha1(C.read()).hexdigest()
	except Exception as D:LOG.warning(f"Failed to open file and compute hash for file at {A}: {D}")
def compute_version_archive_hash(version_no,state_archive):
	def C(**A):B=A.get('dir_name');C=A.get('fname');D=A.get('mutables')[0];E=os.path.join(B,C);compute_file_hash(E,D)
	A=new_tmp_dir()
	with zipfile.ZipFile(state_archive)as D:D.extractall(A)
	E=os.path.join(A,API_STATES_DIR);B=hashlib.sha1();api_states_traverse(E,C,[B]);B.update(str(version_no).encode(_A));rm_rf(A);return B.hexdigest()
def compute_revision_hash(pods_node,obj_store_path):
	A=pods_node
	if not A.state_files:return random_hash()
	D=map(lambda state_file:state_file.hash_ref,A.state_files);B=hashlib.sha1()
	for C in D:
		try:
			with open(os.path.join(obj_store_path,C),'rb')as E:B.update(E.read())
		except Exception as F:LOG.warning(f"Failed to open file and compute hash for {C}: {F}")
	if isinstance(A,Revision):B.update(A.rid.encode(_A));B.update(str(A.revision_number).encode(_A))
	elif isinstance(A,Version):B.update(str(A.version_number).encode(_A))
	return B.hexdigest()


_A='AUTH'
import base64,hmac,smtplib
from base64 import b64encode
from smtplib import SMTPAuthenticationError,SMTPException
from localstack.utils.common import to_bytes,to_str
def base64_encode(s):A=to_str(b64encode(to_bytes(s))).strip();return A
class SMTP(smtplib.SMTP):
	def login(C,user,password):
		L='auth';G=password;F=user
		def M(challenge,user,password):A=challenge;A=base64.decodestring(A);B=user+' '+hmac.HMAC(password,A).hexdigest();return base64_encode(B)
		def N(user,password):return base64_encode('\x00%s\x00%s'%(user,password))
		H='PLAIN';I='CRAM-MD5';J='LOGIN';C.ehlo_or_helo_if_needed()
		if not C.has_extn(L):raise SMTPException('SMTP AUTH extension not supported by server.')
		O=C.esmtp_features[L].split();P=[I,H,J];E=None
		for K in P:
			if K in O:E=K;break
		if E==I:
			A=C.docmd(_A,I);B=A[0];D=A[1]
			if B==503:return B,D
			A=C.docmd(M(D,F,G));B=A[0];D=A[1]
		elif E==H:A=C.docmd(_A,H+' '+N(F,G));B=A[0];D=A[1]
		elif E==J:Q=base64_encode(F);R=base64_encode(G);A=C.check_codes(Q,R);B=A[0];D=A[1]
		elif E is None:raise SMTPException('No suitable authentication method found.')
		if B not in(235,503):raise SMTPAuthenticationError(B,D)
		return B,D
	def check_codes(D,encoded_username,encoded_password):
		E='LOGIN';B=D.docmd(_A,E);A=B[0];C=B[1]
		if A!=334:raise SMTPAuthenticationError(A,C)
		B=D.docmd(encoded_username);A=B[0];C=B[1]
		if A!=334:raise SMTPAuthenticationError(A,C)
		B=D.docmd(encoded_password);A=B[0];C=B[1]
		if A!=235:raise SMTPAuthenticationError(A,C)
		return A,C


_O='--reset/--no-reset'
_N='--inject/--no-inject'
_M='Injects the state from a version into the application runtime'
_L='--remote'
_K='--message'
_J='--version'
_I='-v'
_H=False
_G='name'
_F='Name of the cloud pod'
_E='cpvcs'
_D='backend'
_C='--name'
_B='-n'
_A=True
import json,sys
from typing import Any,List,Mapping,Tuple
import click
from click import Context
from localstack.cli import console
from localstack.utils.analytics.cli import publish_invocation
def required_if_not_cached(option_key):
	class PodConfigContext(click.Option):
		def handle_parse_result(self,ctx,opts,args):
			from localstack_ext.bootstrap import pods_client;is_present=self.name in opts
			if not is_present:
				config_cache=pods_client.get_pods_config();pod_name=config_cache.get(option_key)
				if pod_name is None:raise click.MissingParameter(f"Parameter `--{option_key}` unspecified. Call with `--{option_key}` or set the parameter with `set-context`")
				opts[self.name]=pod_name
			return super().handle_parse_result(ctx,opts,args)
	return PodConfigContext
def _cloud_pod_initialized(pod_name):
	from localstack_ext.bootstrap import pods_client
	if not pods_client.is_initialized(pod_name=pod_name):console.print('[red]Error:[/red] Could not find local CloudPods instance');return _H
	return _A
@click.group(name='pod',help='Cloud Pods with elaborate versioning mechanism')
def pod():
	from localstack_ext.bootstrap.licensing import is_logged_in
	if not is_logged_in():console.print('[red]Error:[/red] not logged in, please log in first');sys.exit(1)
@pod.command(name='set-context',help='Sets the context for all the pod commands')
@click.option(_B,_C,help='Name of the cloud pod to set in the context',required=_A)
@publish_invocation
def cmd_pod_set_context(name):from localstack_ext.bootstrap import pods_client;options=dict(locals());del options['pods_client'];pods_client.save_pods_config(options=options)
@pod.command(name='delete',help='Deletes the specified cloud pod. By default only locally')
@click.option(_B,_C,help=_F,cls=required_if_not_cached(_G))
@click.option('-r',_L,help='Whether the Pod should also be deleted remotely.',is_flag=_A,default=_H)
@publish_invocation
def cmd_pod_delete(name,remote):
	from localstack_ext.bootstrap import pods_client;result=pods_client.delete_pod(pod_name=name,remote=remote,pre_config={_D:_E})
	if result:console.print(f"Successfully deleted {name}")
	else:console.print(f"[yellow]{name} not available locally[/yellow]")
@pod.command(name='rename',help='Renames the pod. If the pod is remotely registered, change is also propagated to remote')
@click.option(_B,_C,help='Current Name of the cloud pod',required=_A)
@click.option('-nn','--new-name',help='New name of the cloud pod',required=_A)
@publish_invocation
def cmd_pod_rename(name,new_name):
	from localstack_ext.bootstrap import pods_client
	if _cloud_pod_initialized(pod_name=name):
		result=pods_client.rename_pod(current_pod_name=name,new_pod_name=new_name,pre_config={_D:_E})
		if result:console.print(f"Successfully renamed {name} to {new_name}")
		else:console.print(f"[red]Error:[/red] Failed to rename {name} to {new_name}")
@pod.command(name='commit',help='Commits the current expansion point and creates a new (empty) revision')
@click.option('-m',_K,help='Add a comment describing the revision')
@click.option(_B,_C,help=_F,cls=required_if_not_cached(_G))
@publish_invocation
def cmd_pod_commit(message,name):from localstack_ext.bootstrap import pods_client;pods_client.commit_state(pod_name=name,pre_config={_D:_E},message=message);console.print('Successfully committed the current state')
@pod.command(name='push',help='Creates a new version by using the state files in the current expansion point (latest commit)')
@click.option('--register/--no-register',default=_A,help='Registers a local Cloud Pod instance with platform')
@click.option('--three-way',is_flag=_A,default=_H,help='')
@click.option('-m',_K,help='Add a comment describing the version')
@click.option(_B,_C,help=_F,cls=required_if_not_cached(_G))
@publish_invocation
def cmd_pod_push(message,name,register,three_way):
	from localstack_ext.bootstrap import pods_client;result=pods_client.push_state(pod_name=name,pre_config={_D:_E},comment=message,register=register);console.print('Successfully pushed the current state')
	if register:
		if result:console.print(f"Successfully registered {name} with remote!")
		else:console.print(f"[red]Error:[/red] Pod with name {name} is already registered")
@pod.command(name='push-overwrite',help='Overwrites a version with the content from the latest commit of the currently selected version')
@click.option(_B,_C,help=_F,cls=required_if_not_cached(_G))
@click.option(_I,_J,type=int)
@click.option('-m',_K,required=_H)
@publish_invocation
def cmd_pod_push_overwrite(version,message,name):
	from localstack_ext.bootstrap import pods_client
	if _cloud_pod_initialized(pod_name=name):
		result=pods_client.push_overwrite(version=version,pod_name=name,comment=message,pre_config={_D:_E})
		if result:console.print('Successfully overwritten state of version ')
@pod.command(name='inject',help=_M)
@click.option(_I,_J,default='-1',type=int,help='Loads the state of the specified version - Most recent one by default')
@click.option('--reset',is_flag=_A,default=_H,help='Will reset the application state before injecting')
@click.option(_B,_C,help=_F,cls=required_if_not_cached(_G))
@publish_invocation
def cmd_pod_inject(version,reset,name):
	from localstack_ext.bootstrap import pods_client;result=pods_client.inject_state(pod_name=name,version=version,reset_state=reset,pre_config={_D:_E})
	if result:console.print('[green]Successfully Injected Pod State[/green]')
	else:console.print('[red]Failed to Inject Pod State[/red]')
@click.option(_N,default=_A,help='Whether the latest version of the pulled pod should be injected')
@click.option(_O,default=_A,help='Whether the current application state should be reset after the pod has been pulled')
@click.option(_B,_C,help=_F,cls=required_if_not_cached(_G))
@click.option('--lazy/--eager',default=_A,help='Will only fetch references to existing versions, i.e. version state is only downloaded when required')
@pod.command(name='pull',help=_M)
@publish_invocation
def cmd_pod_pull(name,inject,reset,lazy):from localstack_ext.bootstrap import pods_client;pods_client.pull_state(pod_name=name,inject_version_state=inject,reset_state_before=reset,lazy=lazy,pre_config={_D:_E})
@pod.command(name='list',help='Lists all pods and indicates which pods exist locally and, by default, which ones are managed remotely')
@click.option(_L,'-r',is_flag=_A,default=_H)
@publish_invocation
def cmd_pod_list_pods(remote):
	from localstack_ext.bootstrap import pods_client;pods=pods_client.list_pods(remote=remote,pre_config={_D:_E})
	if not pods:console.print(f"[yellow]No pods available {'locally'if not remote else''}[/yellow]")
	else:console.print('\n'.join(pods))
@pod.command(name='versions',help='Lists all available version numbers')
@click.option(_B,_C,help=_F,cls=required_if_not_cached(_G))
@publish_invocation
def cmd_pod_versions(name):
	if _cloud_pod_initialized(pod_name=name):from localstack_ext.bootstrap import pods_client;version_list=pods_client.list_versions(pod_name=name,pre_config={_D:_E});result='\n'.join(version_list);console.print(result)
@pod.command(name='version-info')
@click.option(_I,_J,required=_A,type=int)
@click.option(_B,_C,help=_F,cls=required_if_not_cached(_G))
@publish_invocation
def cmd_pod_version_info(version,name):
	if _cloud_pod_initialized(pod_name=name):from localstack_ext.bootstrap import pods_client;info=pods_client.get_version_info(version=version,pod_name=name,pre_config={_D:_E});console.print_json(json.dumps(info))
@pod.command(name='metamodel',help='Displays the content metamodel as json')
@click.option(_I,_J,type=int,default=-1,help='Latest version by default')
@click.option(_B,_C,help=_F,cls=required_if_not_cached(_G))
@publish_invocation
def cmd_pod_version_metamodel(version,name):
	if _cloud_pod_initialized(pod_name=name):
		from localstack_ext.bootstrap import pods_client;metamodel=pods_client.get_version_metamodel(version=version,pod_name=name,pre_config={_D:_E})
		if metamodel:console.print_json(json.dumps(metamodel))
		else:console.print(f"[red]Could not find metamodel for pod {name} with version {version}[/red]")
@pod.command(name='set-version',help='Set HEAD to a specific version')
@click.option(_I,_J,required=_A,type=int,help='The version the state should be set to')
@click.option(_N,default=_A,help='Whether the state should be directly injected into the application runtime after changing version')
@click.option(_O,default=_A,help='Whether the current application state should be reset before changing version')
@click.option('--commit-before',is_flag=_H,help='Whether the current application state should be committed to the currently selected version before changing version')
@click.option(_B,_C,help=_F,cls=required_if_not_cached(_G))
@publish_invocation
def cmd_pod_set_version(version,inject,reset,commit_before,name):
	if _cloud_pod_initialized(pod_name=name):from localstack_ext.bootstrap import pods_client;pods_client.set_version(version=version,inject_version_state=inject,reset_state=reset,commit_before=commit_before,pod_name=name,pre_config={_D:_E})
@pod.command(name='commits',help='Shows the commit history of a version')
@click.option(_J,_I,default=-1)
@click.option(_B,_C,help=_F,cls=required_if_not_cached(_G))
@publish_invocation
def cmd_pod_commits(version,name):
	if _cloud_pod_initialized(pod_name=name):from localstack_ext.bootstrap import pods_client;commits=pods_client.list_version_commits(version=version,pod_name=name,pre_config={_D:_E});result='\n'.join(commits);console.print(result)
@pod.command(name='commit-diff',help='Shows the changes made by a commit')
@click.option(_J,_I,required=_A)
@click.option('--commit','-c',required=_A)
@click.option(_B,_C,help=_F,cls=required_if_not_cached(_G))
@publish_invocation
def cmd_pod_commit_diff(version,commit,name):
	if _cloud_pod_initialized(pod_name=name):
		from localstack_ext.bootstrap import pods_client;commit_diff=pods_client.get_commit_diff(version=version,commit=commit,pod_name=name,pre_config={_D:_E})
		if commit_diff:console.print_json(json.dumps(commit_diff))
		else:console.print(f"[red]Error:[/red] Commit {commit} not found for version {version}")